@@ -1,3 +1,4 @@
+from dataclasses import dataclass
 from typing import Dict, TypeVar, Union, NamedTuple, Optional, Any, Callable, List, Iterable, Tuple, Mapping, Sequence
 import functools
 import sys
@@ -12,7 +13,8 @@ from plover_stroke import BaseStroke  # type: ignore
 
 #from plover.system import english_stenotype as system
 
-Context=SimpleNamespace
+
+
 
 def concatenate_merge_function(*args, **kwargs)->str:
 	return "".join(*args, *kwargs)
@@ -113,6 +115,8 @@ class Dictionary:
 		"""
 		Compute the union of two dictionaries.
 		"""
+		if self.longest_key==0: return other
+		if other.longest_key==0: return self
 		return AlternativeDictionary(self.stroke_type, [self, other])
 
 	def __mul__(self, other: "Dictionary")->"Dictionary":
@@ -130,10 +134,18 @@ class Dictionary:
 	def map(self, function: Callable[..., Any])->"Dictionary":
 		"""
 		Map a function over the dictionary values.
+
+		The function can return None or raise KeyError to signify "no entry".
+
+		Note that any empty value is interpreted by Plover as "no entry". The no-op translation
+		in Plover is {#}.
 		"""
 		return MappedDictionary(self.stroke_type, self, function)
 
-	def filter(self, condition: Callable[[Strokes, Any], Any])->"Dictionary":
+	def filter(self, condition: Callable[..., Any])->"Dictionary":
+		"""
+		See apply_function for the possible signature of condition.
+		"""
 		return FilteredDictionary(self.stroke_type, self, condition)
 
 	def named(self, name: str)->"NamedDictionary":
@@ -292,7 +304,11 @@ class SingleDictionary(Dictionary):
 		else:
 			assert False
 
-		assert self.data
+		if not self.data:
+			self.outline_length=0
+			self.longest_key=0
+			return
+
 		self.outline_length=len(next(iter(self.data.keys())))
 		self.longest_key=max(len(strokes) for strokes in self.data.keys())
 		if any(len(strokes)!=self.outline_length for strokes in self.data.keys()):
@@ -309,6 +325,18 @@ class SingleDictionary(Dictionary):
 	def items(self)->Iterable[Tuple[Strokes, Any]]:
 		return self.data.items()
 
+	def __iter__(self)->Iterable[Strokes]:
+		return self.data.__iter__()
+
+	def keys(self)->Iterable[Strokes]:
+		return self.data.keys()
+
+	def values(self)->Iterable[Any]:
+		return self.data.values()
+
+	def __contains__(self, strokes: Strokes)->bool:
+		return self.data.__contains__(strokes)
+
 
 class ProductDictionary(Dictionary):
 	def __init__(self, stroke_type: type, a: Dictionary, b: Dictionary, merge: bool)->None:
@@ -467,6 +495,36 @@ def translation(stroke_type: type, translation: str)->Dictionary:
 	"""
 	return SingleDictionary(stroke_type, {"": translation})
 
+
+
+# define type alias for mypy type checking below
+SingleDictionary_=SingleDictionary
+FilteredDictionary_=FilteredDictionary
+SubsetDictionary_=SubsetDictionary
+
+
+@dataclass
+class Context:
+	# first argument is self. Mypy currently doesn't type-check correctly
+	Stroke            : Callable[[Any,    InputStrokeType], BaseStroke]
+	stroke_type       : Callable[[Any,    InputStrokeType], BaseStroke]
+	SingleDictionary  : Callable[[Any,    InputStrokeType], SingleDictionary_]
+	s                 : Callable[[Any,    InputStrokeType], SingleDictionary_]
+
+	stroke            : Callable[[Any,    str], Dictionary]  # consist of that single stroke. See example file for details
+
+	translation       : Callable[[Any,    Any], Dictionary]  # consist of a "null stroke" that translates to the translation. Input type is most frequently str
+
+	filtered          : Callable[[Any,    Any], FilteredDictionary_]
+	FilteredDictionary: Callable[[Any,    Any], FilteredDictionary_]
+
+	subsets           : Callable[[Any,    Any], SubsetDictionary_]
+	subsetd           : Callable[[Any,    Any], SubsetDictionary_]
+	SubsetDictionary  : Callable[[Any,    Any], SubsetDictionary_]
+
+
+
+
 def get_context(stroke_type: type)->Context:
 	return Context(
 			Stroke            =stroke_type,
@@ -488,5 +546,7 @@ def get_context(stroke_type: type)->Context:
 
 def get_context_from_system(system: Any)->Context:
 	class Stroke_(BaseStroke): pass
-	Stroke_.setup(system.KEYS, system.IMPLICIT_HYPHEN_KEYS, system.NUMBER_KEY, system.NUMBERS)
+	Stroke_.setup(system.KEYS,
+			{*system.IMPLICIT_HYPHEN_KEYS} & {*system.KEYS},
+			system.NUMBER_KEY, system.NUMBERS)
 	return get_context(Stroke_)
